"""
An interface for implementing a temperature controller device driver for the CETONI SiLA SDK
"""

from __future__ import annotations

import datetime
import logging
import math
import time
import traceback
from abc import abstractmethod, abstractstaticmethod
from dataclasses import dataclass
from typing import List, NamedTuple, Optional

import serial
import serial.threaded
import serial.tools.list_ports
from sila_cetoni.device_driver_abc import DeviceDriverABC

from .utils import StrEnum

logger = logging.getLogger(__name__)


class _ReaderThread(serial.threaded.ReaderThread):
    """
    Implements a serial port read loop and dispatches to a Protocol instance but do it with threads.

    Provides the connection from the `SerialTemperatureControllerInterface` instance (which creates an instance of this
    `_ReaderThread` class) to the protocol instance (`_SerialTemperatureControllerReader`) that is created by this
    `_ReaderThread` so that the protocol instance that receives the data from the serial port can set the actual value
    of the temperature controller in the `SerialTemperatureControllerInterface` instance
    """

    temp_controller: SerialTemperatureControllerInterface = None

    def __init__(self, serial_instance, protocol_factory, temp_controller):
        super().__init__(serial_instance, protocol_factory)
        self.temp_controller: SerialTemperatureControllerInterface = temp_controller


class _SerialTemperatureControllerReader(serial.threaded.LineReader):
    """
     Reads/writes to/from a temperature controller via a serial connection

     The `SerialTemperatureControllerInterface` instance that is represented is obtained from the transport object (a
    `_ReaderThread`) that creates an instance of this reader class
    """

    __temp_controller: SerialTemperatureControllerInterface = None

    # implements serial.threaded.LineReader ----------------------------------
    def connection_made(self, transport: _ReaderThread):
        super().connection_made(transport)

        self.__temp_controller: SerialTemperatureControllerInterface = transport.temp_controller
        logger.debug("port opened")

        # validate that we have the correct temperature controller
        self.write_line(self.__temp_controller.unique_temperature_controller_identifier_request)
        time.sleep(1.5)
        buffer: bytes = transport.serial.read_all()
        logger.debug(f"buffer {buffer}")
        if not buffer or not self.__temp_controller.is_valid_temperature_controller(buffer):
            raise self.__temp_controller.not_found_exception()

    def handle_line(self, data: str):
        logger.debug("line received: {}".format(repr(data)))

        self.__temp_controller.on_serial_data_received(data)

    def connection_lost(self, exc):
        if exc:
            traceback.print_exception(exc, exc, None)
        logger.debug("port closed")


class RangeLimit(NamedTuple):
    """
    Simple class to hold a minimum and maximum range limit
    """

    min: float = -1
    max: float = -1


class TemperatureControlModeInterface(StrEnum):
    """
    Enum interface for the different temperature control modes

    Derive from this class to specify the device-specific control modes
    """

    pass


class TemperatureControllerInterface(DeviceDriverABC):
    """
    Interface for a temperature controller device driver
    """

    _name: str

    def __init__(self):
        super().__init__()

        self._name = ""

    @property
    def name(self) -> str:
        """
        Returns the name of the temperature controller
        """
        return self._name


class HuberChillerInterface(TemperatureControllerInterface):
    """
    Interface for all Huber Chiller device drivers
    """

    _operating_range_limit: RangeLimit
    _set_point_range_limit: RangeLimit
    _set_point_value: float
    _internal_actual_value: float
    _external_actual_value: float
    _temperature_control_mode: TemperatureControlModeInterface
    _alarm_status: bool

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._operating_range_limit = RangeLimit()
        self._set_point_range_limit = RangeLimit()
        self._set_point_value = 0.0
        self._internal_actual_value = 0.0
        self._external_actual_value = 0.0
        self._temperature_control_mode = None
        self._alarm_status = False

    @property
    def operating_range_limit(self) -> RangeLimit:
        """
        Returns the current operating range limit of the temperature controller
        """
        return self._operating_range_limit

    @property
    def set_point_range_limit(self) -> RangeLimit:
        """
        Returns the current set point range limit of the temperature controller
        """
        return self._set_point_range_limit

    @property
    def set_point_value(self) -> float:
        """
        Returns the current set point value of the temperature controller
        """
        return self._set_point_value

    @set_point_value.setter
    @abstractmethod
    def set_point_value(self, set_point: float) -> None:
        """
        Set the set point value of the temperature controller
        """
        raise NotImplementedError()

    @property
    def internal_actual_value(self) -> float:
        """
        Returns the current internal actual value of the temperature controller
        """
        return self._internal_actual_value

    @property
    def external_actual_value(self) -> float:
        """
        Returns the current external actual value of the temperature controller
        """
        return self._external_actual_value

    @property
    def temperature_control_mode(self) -> TemperatureControlModeInterface:
        """
        Returns the current temperature control mode of the temperature controller
        """
        return self._temperature_control_mode

    @temperature_control_mode.setter
    @abstractmethod
    def temperature_control_mode(self, mode: TemperatureControlModeInterface) -> None:
        """
        Set the temperature control mode of the temperature controller
        """
        raise NotImplementedError()

    @property
    def alarm_status(self) -> bool:
        """
        Returns the current alarm status of the temperature controller

        `True` means the device is in an alarm state, `False` means it is not in an alarm state
        """
        return self._alarm_status

    @abstractmethod
    def clear_alarm_status(self) -> None:
        """
        Clears the alarm state of the temperature controller
        """
        raise NotImplementedError()


@dataclass
class Range:
    """
    Simple class to hold a minimum and maximum range
    """

    min: float = -1
    max: float = -1

    def __eq__(self, o: Range) -> bool:
        return math.isclose(self.min, o.min) and math.isclose(self.max, o.max)

    def __getitem__(self, key) -> float:
        """
        This allows iterating over the fields of this class and allows to use the star expression to unpack the fields
        for use as function parameters, for example

            >>> r = Range(1, 100)
            >>> print(*r)
            1 100
        """
        if key == 0:
            return self.min
        if key == 1:
            return self.max
        raise IndexError


class EventLogEntryType(StrEnum):
    """
    The possible types of an event log entry
    """

    INFORMATION = "i"
    EVENT_OCCURRED = "+"
    EVENT_FINISHED = "-"


class EventLogEntry(NamedTuple):
    """
    A simple event log entry class
    """

    timestamp: datetime.datetime
    entry_type: EventLogEntryType
    code: int
    description: str


class MemmertOvenInterface(TemperatureControllerInterface):
    """
    Interface for all Memmert Oven device drivers
    """

    _set_point_value: float
    _actual_temperature: float
    _alarm_temperature: Range
    _set_point_range: Range
    _alarm_temperature_range: Range
    _fan_rotation_speed: int
    _flap_position: int
    _min_temperature_underrun: bool
    _max_temperature_exceeded: bool
    _event_log: List[EventLogEntry]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._set_point_value = 0.0
        self._actual_temperature = 0.0
        self._alarm_temperature = Range()
        self._set_point_range = Range()
        self._alarm_temperature_range = Range()
        self._fan_rotation_speed = 0
        self._flap_position = 0
        self._min_temperature_underrun = False
        self._max_temperature_exceeded = False
        self._event_log = []

    @property
    def set_point_value(self) -> float:
        """
        Returns the current set point value of the temperature controller
        """
        return self._set_point_value

    @set_point_value.setter
    @abstractmethod
    def set_point_value(self, set_point: float) -> None:
        """
        Set the set point value of the temperature controller
        """
        raise NotImplementedError()

    @property
    def actual_temperature(self) -> float:
        """
        Returns the current actual temperature of the temperature controller
        """
        return self._actual_temperature

    @property
    def alarm_temperature(self) -> Range:
        """
        Returns the current alarm temperature
        """
        return self._alarm_temperature

    @alarm_temperature.setter
    @abstractmethod
    def alarm_temperature(self, alarm_temperature: Range) -> None:
        """
        Set the alarm temperatures(minimum and maximum)
        """
        raise NotImplementedError()

    def set_alarm_temperature_min(self, alarm_temperature_min: float) -> None:
        """
        Set the minimum alarm temperature
        """
        raise NotImplementedError

    def set_alarm_temperature_max(self, alarm_temperature_max: float) -> None:
        """
        Set the maximum alarm temperature
        """
        raise NotImplementedError

    @property
    def set_point_range(self) -> Range:
        """
        Returns the current set point range
        """
        return self._set_point_range

    @property
    def alarm_temperature_range(self) -> Range:
        """
        Returns the current alarm temperature range
        """
        return self._alarm_temperature_range

    @property
    def fan_rotation_speed(self) -> float:
        """
        Returns the current fan rotation speed
        """
        return self._fan_rotation_speed

    @fan_rotation_speed.setter
    @abstractmethod
    def fan_rotation_speed(self, fan_rotation_speed: int) -> None:
        """
        Set the fan rotation speed
        """
        raise NotImplementedError()

    @property
    def flap_position(self) -> int:
        """
        Returns the current flap position
        """
        return self._flap_position

    @flap_position.setter
    @abstractmethod
    def flap_position(self, flap_position: int) -> None:
        """
        Set the flap position
        """
        raise NotImplementedError()

    @property
    def min_temperature_underrun(self) -> bool:
        """
        Returns whether the current temperature is lower than the minimum temperature
        """
        return self._min_temperature_underrun

    @property
    def max_temperature_exceeded(self) -> bool:
        """
        Returns whether the current temperature is higher than the maximum temperature
        """
        return self._max_temperature_exceeded

    @property
    def event_log(self) -> List[EventLogEntry]:
        """
        Returns the event log
        """
        return self._event_log


class TemperatureControllerNotFoundException(Exception):
    """
    An exception that indicates that a temperature controller is not available (any more) over a serial connection
    """

    def __init__(self, port: Optional[str] = None, msg: str = "No temperature controller detected on {port}"):
        super().__init__(msg.format(port=port if port is not None else "any serial port"))


class SerialTemperatureControllerInterface(TemperatureControllerInterface):
    """
    Interface for a temperature controller device driver that uses a serial communication protocol
    """

    # You can derive from `TemperatureControllerNotFoundException` to give a more detailed error
    # message. Set `not_found_exception = YourTemperatureControllerNotFoundException` in your
    # custom TemperatureController class
    not_found_exception = TemperatureControllerNotFoundException
    # The delimiter that is sent at the end of every message
    delimiter: str = "\r"

    __serial: serial.Serial
    __reader_thread: serial.threaded.ReaderThread
    _protocol: _SerialTemperatureControllerReader

    logger = logging.getLogger(__name__)

    def __init__(self, port: Optional[str] = None):
        """
        :param port: (optional) The serial port of the temperature controller
        """
        super().__init__()

        try:
            self.__serial = serial.Serial(
                port,
                baudrate=115200,
                bytesize=serial.EIGHTBITS,
                parity=serial.PARITY_NONE,
                stopbits=serial.STOPBITS_ONE,
            )
        except serial.SerialException:
            raise self.not_found_exception(port)
        self.__reader_thread = _ReaderThread(self.__serial, _SerialTemperatureControllerReader, self)
        self.__reader_thread.name = "SerialTemperatureControllerReaderThread"

    def _autodetect_serial_port(self):
        """
        Goes through all available serial ports of the system and tries to find the first port connected to a
        temperature controller
        """

        for info in serial.tools.list_ports.comports():
            logger.debug(f"autodetection trying port {info.device}")
            try:
                ser = serial.Serial(
                    info.device,
                    baudrate=115200,
                    bytesize=serial.EIGHTBITS,
                    parity=serial.PARITY_NONE,
                    stopbits=serial.STOPBITS_ONE,
                    timeout=2,
                    write_timeout=2,
                )
                # read something from the temperature controller that uniquely identifies it
                TERMINATOR = _SerialTemperatureControllerReader.TERMINATOR
                s = f"{self.unique_temperature_controller_identifier_request}".encode("utf-8") + TERMINATOR
                logger.debug(f"send {s}")
                ser.write(f"{self.unique_temperature_controller_identifier_request}".encode("utf-8") + TERMINATOR)
                time.sleep(0.1)
                if self.is_valid_temperature_controller(ser.read_all()):
                    logger.info(f"Temperature Controller detected on serial port {ser.port}")
                    self.__serial.port = ser.port
                    break
            except (TimeoutError, serial.SerialTimeoutException, serial.SerialException):
                continue
        else:
            raise self.not_found_exception()

    def open(self, port: str = ""):
        """
        Connects to the temperature controller via the serial port with the given `port`

        :param port: (optional) The serial port of the temperature controller. If not given, tries to autodetect a
                     temperature controller by trying all available serial ports until a valid port is found
        """
        if self.__serial.isOpen():
            return

        if not port:
            self._autodetect_serial_port()
        else:
            self.__serial.port = port

        self.__serial.open()

    def start(self):
        if not self.__serial.isOpen():
            raise self.not_found_exception(self.port)
        try:
            self.__reader_thread.start()
            self._protocol: _SerialTemperatureControllerReader = self.__reader_thread.connect()[1]
        except RuntimeError:
            raise self.not_found_exception(self.port)

    def stop(self):
        self.close()

    def close(self):
        """
        Closes the communication to the temperature controller
        """
        if self.__reader_thread.is_alive():
            self.__reader_thread.close()
        self.__serial.close()
        logger.debug("stop done")

    @property
    def port(self) -> str:
        """
        Returns the serial port through which this device driver communicates with the temperature controller
        """
        return self.__serial.port

    @abstractmethod
    def on_serial_data_received(self, data: str) -> None:
        """
        Function to handle the `data` read from the temperature controller
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def unique_temperature_controller_identifier_request(self) -> str:
        """
        This function should return a string containing a request for the temperature controller to return some kind of
        unique identifier or something that can be used to identify if the current serial connection uses the correct
        temperature controller (e.g. this could be a request to return the serial number of the temperature controller).
        The response of the device will be passed to the `is_valid_temperature_controller()` function to validate it.
        """
        raise NotImplementedError()

    @abstractstaticmethod
    def is_valid_temperature_controller(data: bytes) -> bool:
        """
        This function should return whether `data` indicates that the current serial connection is using the correct
        temperature controller.

        :param data: The serial data which was read as a response to writing the
                     `unique_temp_controller_identifier_request`
        """
        raise NotImplementedError()
