"""
A device driver implementation for the serial interface of a Huber Chiller

:author: Florian Meinicke (florian.meinicke@cetoni.de)
:date: 29.06.2022
"""

from __future__ import annotations

import abc
import logging
import math
import random
import time
from queue import Empty, Queue
from threading import Event, Thread
from typing import Optional, Type, Union

from serial import SerialException

from .abc import (
    HuberChillerInterface,
    RangeLimit,
    SerialTemperatureControllerInterface,
    TemperatureControllerNotFoundException,
    TemperatureControlModeInterface,
    _SerialTemperatureControllerReader,
)
from .utils import StrEnum, hex_to_int, hex_to_temperature, int_to_hex, temperature_to_hex

TERMINATOR = "\r"  # terminator as per manual
_SerialTemperatureControllerReader.TERMINATOR = TERMINATOR.encode("utf-8")

logger = logging.getLogger(__name__)


class HuberTemperatureControlMode(TemperatureControlModeInterface):
    """
    The temperature control modes for Huber chillers
    """

    CIRCULATION = "C"  #: only circulation, no actual cooling/heating
    INTERNAL = "I"  #: circulation and cooling/heating
    OFF = "O"  #: neither circulation nor cooling/heating
    NO_CHANGE = "*"


class HuberAlarmAcknowledgement(StrEnum):
    """
    The ways to acknowledge an alarm of a Huber chiller
    """

    NO_ACKNOWLEDGEMENT = "0"
    ACKNOWLEDGEMENT = "1"
    NO_CHANGE = "*"


class _LAICommandType(StrEnum):
    VERIFY = "V"
    LIMIT = "L"
    GENERAL = "G"


class _LAIIdentifier(StrEnum):
    MASTER = "M"
    SLAVE = "S"


class _LAIBase(abc.ABC):
    """
    Base class for Huber's LAI protocol
    """

    _start_char: str = "["
    _identifier: _LAIIdentifier
    _slave_address: int = 1  # fixed, as per manual
    _command_type: _LAICommandType
    _length: int
    _payload: str
    _checksum: str = ""  # 1-byte hexadecimal sum of the previous fields

    # 1 byte for the start char '['
    # 1 byte for the master/slave identifier 'M' or 'S'
    # 2 bytes for the slave address '01'
    # 1 byte for the command type 'V', 'L', or 'G'
    # 2 bytes for the length of the command
    __MIN_LENGTH: int = 7

    def __init__(self, id: _LAIIdentifier, command_type: _LAICommandType, payload: str = "") -> None:
        self._identifier = id
        self._command_type = command_type
        self._payload = payload
        self._length = self.__MIN_LENGTH + len(payload)
        self.__calc_checksum()

    @classmethod
    def from_str(
        cls: Union[Type[_LAICommand], Type[_LAIResponse]], value: str
    ) -> Optional[Union[_LAICommand, _LAIResponse]]:
        """
        Converts the given string `value` to an instance of `cls`

        Validates that `value` is in the correct LAI format and that the checksums match. If either is not given this
        function returns `None`.
        """

        if cls not in (_LAICommand, _LAIResponse):
            logger.error(f"Expected either '_LAICommand' or '_LAIResponse' in 'from_str' but got {cls!r}.")
            return None

        if len(value) < _LAIBase.__MIN_LENGTH:
            logger.error(
                f"Incorrect format of LAI string! The string must have a length of at least {_LAIBase.__MIN_LENGTH}. "
                f"Got {value!r}"
            )
            return None

        if value[0] != cls._start_char:
            logger.error(
                f"Incorrect format of LAI string! Expected {cls._start_char!r} at position 0 but got {value[0]!r}."
            )
            return None

        expected_id = _LAIIdentifier.MASTER if cls == _LAICommand else _LAIIdentifier.SLAVE
        if value[1] != expected_id:
            logger.error(
                f"Unexpected identifier in LAI string! Expected {expected_id.value!r} at position 1 but got "
                f"{value[1]!r}."
            )
            return None

        slave_address = hex_to_int(value[2:4])
        if slave_address != cls._slave_address:
            logger.error(
                f"Unexpected slave address in LAI string! Expected {cls._slave_address!r} but got {slave_address!r}."
            )
            return None

        length = hex_to_int(value[5:7])
        payload = value[7:-2]
        if length != _LAIBase.__MIN_LENGTH + len(payload):
            logger.warning(
                f"Incorrect length of LAI string! Expected {_LAIBase.__MIN_LENGTH + len(payload)!r} but got {length!r}."
            )

        obj = cls(_LAICommandType(value[4]), payload)

        if obj.checksum != value[-2:]:
            logger.error(f"Checksums don't match! Expected {obj.checksum!r} but got {value[-2:]!r}")
            return None

        EXPECTED_PAYLOAD_LENGTHS = {_LAICommandType.VERIFY: 13, _LAICommandType.LIMIT: 16, _LAICommandType.GENERAL: 14}
        expected_payload_length = EXPECTED_PAYLOAD_LENGTHS[obj.command_type]
        if len(obj.payload) != expected_payload_length:
            logger.error(
                f"Unexpected response for the {obj.command_type} command! Expected payload to be "
                f"{expected_payload_length} bytes. Got {resp!r}."
            )
            return None

        return obj

    @property
    def command_type(self) -> str:
        return self._command_type

    @property
    def payload(self) -> str:
        return self._payload

    @property
    def checksum(self) -> str:
        return self._checksum

    def __calc_checksum(self) -> int:
        """
        Calculates the command's checksum and stores it in `__checksum`
        """
        checksum = 0
        for c in str(self):
            checksum += ord(c)
        self._checksum = int_to_hex(checksum)

    def __str__(self) -> str:
        """
        The command's string representation that can be send via a serial connection
        """
        return (
            self._start_char
            + self._identifier
            + int_to_hex(self._slave_address)
            + self._command_type
            + int_to_hex(self._length)
            + self._payload
            + self._checksum
        )

    def __repr__(self) -> str:
        return f"_LAIBase({self._identifier!r}, {self._command_type!r}, {self._payload!r})"

    def __eq__(self, other):
        return repr(self) == repr(other)


class _LAICommand(_LAIBase):
    """
    Huber's LAI commands from the master
    """

    VERIFY: _LAICommand  #: the VERIFY command
    LIMIT: _LAICommand  #: the LIMIT command to get the operating and set point range limits
    GENERAL: _LAICommand  #: the GENERAL command to get the current set point and actual values

    def __init__(
        self,
        command_type: _LAICommandType,
        *,
        payload: Optional[str] = None,
        temp_control_mode: HuberTemperatureControlMode = HuberTemperatureControlMode.NO_CHANGE,
        clear_alarm: HuberAlarmAcknowledgement = HuberAlarmAcknowledgement.NO_CHANGE,
        set_point: Optional[float] = None,
    ) -> None:
        if payload is not None and set_point is not None:
            logger.error(
                "Cannot specify 'payload' along with 'temp_control_mode', 'clear_alarm', and 'set_point'! Use either "
                "'payload' or specify all of 'temp_control_mode', 'clear_alarm', and 'set_point'."
            )
            super().__init__(_LAIIdentifier.MASTER, command_type)
            return
        if payload is None:
            payload = (
                ""
                if command_type == _LAICommandType.VERIFY
                else temp_control_mode
                + clear_alarm
                + (temperature_to_hex(set_point) if set_point is not None else "****")
            )
        super().__init__(
            _LAIIdentifier.MASTER,
            command_type,
            payload,
        )

    def __repr__(self) -> str:
        return f"_LAICommand({self._command_type!r}, {self._payload!r})"


_LAICommand.VERIFY = _LAICommand(_LAICommandType.VERIFY)
_LAICommand.LIMIT = _LAICommand(_LAICommandType.LIMIT, payload="********")
_LAICommand.GENERAL = _LAICommand(_LAICommandType.GENERAL)


class _LAIResponse(_LAIBase):
    """
    Huber's LAI responses from the slave
    """

    def __init__(self, command_type: _LAICommandType, payload: str = "") -> None:
        super().__init__(_LAIIdentifier.SLAVE, command_type, payload)

    def __repr__(self) -> str:
        return f"_LAIResponse({self._command_type!r}, {self._payload!r})"


class HuberChillerNotFoundException(TemperatureControllerNotFoundException):
    def __init__(self, port: Optional[str] = None):
        super().__init__(
            port,
            "No Huber chiller detected on {port}. Please verify that your chiller has the following RS232/USB "
            "settings: [Baudrate: 115200], [Modus: HuberBus]. Then restart the device.",
        )


class HuberChiller(HuberChillerInterface, SerialTemperatureControllerInterface):
    """
    Class for reading serial data from a Huber chiller
    """

    not_found_exception = HuberChillerNotFoundException

    __write_queue: Queue[_LAICommand]
    __verify_data_received_event: Event
    __limit_data_received_event: Event
    __general_data_received_event: Event

    def __init__(self, port: Optional[str] = None):
        super().__init__(port)

        self.__write_queue = Queue()
        self.__verify_data_received_event = Event()
        self.__limit_data_received_event = Event()
        self.__general_data_received_event = Event()

    def start(self):
        super().start()

        def poll():
            while self._protocol.transport.alive:
                self.__write_command_and_wait(_LAICommand.LIMIT)
                try:
                    self.__write_command_and_wait(self.__write_queue.get(block=True, timeout=1))
                except Empty:
                    self.__write_command_and_wait(_LAICommand.GENERAL)
                    time.sleep(0.2)

        # fill all members with initial values
        self.__write_command_and_wait(_LAICommand.VERIFY)
        self.__write_command_and_wait(_LAICommand.LIMIT)
        self.__write_command_and_wait(_LAICommand.GENERAL)
        # start polling for value changes
        Thread(target=poll, name="PollingThread").start()

    def __write_command_and_wait(self, command: _LAICommand):
        """
        Writes the given LAI command `command` to the device and waits until the response has been received

        If the response is not received after 1 second it means that either the response has been lost or that the
        device did not send a response. Usually, the response should be received after 300ms (as per device manual).
        """
        WAIT_FOR_RESPONSE_TIMEOUT = 1

        logger.debug(f"Writing {command!r}")
        try:
            self._protocol.write_line(str(command))
            if command.command_type == _LAICommandType.VERIFY:
                if not self.__verify_data_received_event.wait(WAIT_FOR_RESPONSE_TIMEOUT):
                    logger.warning(f"Waiting for VERIFY response timed out after {WAIT_FOR_RESPONSE_TIMEOUT} seconds")
                self.__verify_data_received_event.clear()
            elif command.command_type == _LAICommandType.LIMIT:
                if not self.__limit_data_received_event.wait(WAIT_FOR_RESPONSE_TIMEOUT):
                    logger.warning(f"Waiting for LIMIT response timed out after {WAIT_FOR_RESPONSE_TIMEOUT} seconds")
                self.__limit_data_received_event.clear()
            elif command.command_type == _LAICommandType.GENERAL:
                if not self.__general_data_received_event.wait(WAIT_FOR_RESPONSE_TIMEOUT):
                    logger.warning(f"Waiting for GENERAL response timed out after {WAIT_FOR_RESPONSE_TIMEOUT} seconds")
                self.__general_data_received_event.clear()
        except SerialException as err:
            logger.error(err, exc_info=err)

    def __send_command(self, command: _LAICommand) -> None:
        """
        Sends the given LAI command `command` to the device
        """
        self.__write_queue.put(command)

    @HuberChillerInterface.set_point_value.setter
    def set_point_value(self, set_point: float) -> None:
        """
        Set the set point value of the temperature controller to `set_point`
        """
        logger.debug(f"Setting set point to {set_point}")
        self.__send_command(_LAICommand(_LAICommandType.GENERAL, set_point=set_point))

    @HuberChillerInterface.temperature_control_mode.setter
    def temperature_control_mode(self, mode: Union[str, HuberTemperatureControlMode]) -> None:
        """
        Set teh temperature control mode of the temperature controller
        """
        if isinstance(mode, str):
            if len(mode) > 1:
                mode = HuberTemperatureControlMode[mode.upper()]
            else:
                mode = HuberTemperatureControlMode(mode)
        logger.debug(f"Setting temperature control mode to {mode}")
        self.__send_command(_LAICommand(_LAICommandType.GENERAL, temp_control_mode=mode))

    def clear_alarm_status(self):
        """
        Clears the alarm state of the temperature controller
        """
        logger.debug(f"Clearing alarm state")
        self.__send_command(_LAICommand(_LAICommandType.GENERAL, clear_alarm=HuberAlarmAcknowledgement.ACKNOWLEDGEMENT))

    def on_serial_data_received(self, data: str) -> None:
        """
        Function to handle the `data` read from the temperature controller
        """
        logger.debug(data)
        resp = _LAIResponse.from_str(data[:-1] if data.endswith(TERMINATOR) else data)
        if resp is None:
            logger.warning(f"Invalid LAI response received: {data!r}")
        elif resp.command_type == _LAICommandType.VERIFY:
            self._name = resp.payload
            self.__verify_data_received_event.set()
        elif resp.command_type == _LAICommandType.LIMIT:
            self._set_point_range_limit = RangeLimit(
                hex_to_temperature(resp.payload[0:4]), hex_to_temperature(resp.payload[4:8])
            )
            self._operating_range_limit = RangeLimit(
                hex_to_temperature(resp.payload[8:12]), hex_to_temperature(resp.payload[12:16])
            )
            self.__limit_data_received_event.set()
        elif resp.command_type == _LAICommandType.GENERAL:
            self._temperature_control_mode = HuberTemperatureControlMode(resp.payload[0])
            self._alarm_status = resp.payload[1] == "1"
            self._set_point_value = hex_to_temperature(resp.payload[2:6])
            self._internal_actual_value = hex_to_temperature(resp.payload[6:10])
            self._external_actual_value = hex_to_temperature(resp.payload[10:14])
            self.__general_data_received_event.set()
        else:
            logger.warning(f"Received LAI response with unknown command type {resp.command_type!r}")

    @property
    def unique_temperature_controller_identifier_request(self) -> str:
        return str(_LAICommand.VERIFY)

    @staticmethod
    def is_valid_temperature_controller(data: bytes) -> bool:
        data_str = data.decode()
        if data_str.endswith(TERMINATOR):
            data_str = data_str[:-1]
        logger.debug(f"data {data}, data_str {data_str}")
        return _LAIResponse.from_str(data_str) == _LAIResponse(_LAICommandType.VERIFY, "Huber Control")


class HuberChillerSim(HuberChillerInterface):
    """
    A simple simulated device driver implementation for Huber chillers
    """

    __stop_event: Event

    def __init__(self, *args, **kwargs):
        super().__init__()

        self._name = "Huber Control Sim"
        self._operating_range_limit = RangeLimit(-50, 100)
        self._set_point_range_limit = RangeLimit(5, 50)
        self._set_point_value = 0.0
        self._internal_actual_value = 23.0 + random.random()
        self._external_actual_value = 23.0 + random.random()
        self._temperature_control_mode = HuberTemperatureControlMode.OFF
        self._alarm_status = False
        self.__stop_event = Event()

    def start(self):
        def generate_values():
            while not self.__stop_event.wait(0.5):
                if not self._temperature_control_mode == HuberTemperatureControlMode.OFF:
                    diff = self._set_point_value - self._internal_actual_value
                    self._internal_actual_value += 0.1 * diff * random.random()
                    self._external_actual_value = self._internal_actual_value + random.random()
                    # just to show that alarm status works, set it when temp is around 42
                    self._alarm_status = math.isclose(self._internal_actual_value, 42, abs_tol=0.9)

        Thread(target=generate_values).start()

    def stop(self):
        self.__stop_event.set()

    @HuberChillerInterface.set_point_value.setter
    def set_point_value(self, set_point: float) -> None:
        """
        Set the set point value of the temperature controller to `set_point`
        """
        self._set_point_value = set_point

    @HuberChillerInterface.temperature_control_mode.setter
    def temperature_control_mode(self, mode: Union[str, HuberTemperatureControlMode]) -> None:
        """
        Set teh temperature control mode of the temperature controller
        """
        if isinstance(mode, str):
            if len(mode) > 1:
                self._temperature_control_mode = HuberTemperatureControlMode[mode.upper()]
            else:
                self._temperature_control_mode = HuberTemperatureControlMode(mode)

    def clear_alarm_status(self):
        """
        Clears the alarm state of the temperature controller
        """
        self._alarm_status = False


# ----------------------------------------------------------------------------
# test
if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s [%(threadName)-12.12s] %(levelname)-8s| %(name)s %(module)s.%(funcName)s: %(message)s",
        level=logging.DEBUG,
    )

    print(_LAICommand.VERIFY)  # expect: '[M01V07C6'
    print(_LAICommand.LIMIT)  # expect: '[M01L0F********1B'
    print(_LAICommand.GENERAL)  # expect: '[M01G0D******1B'
    _LAICommand.from_str("[S01V14Huber ControlC1")  # expect error because str is not a _LAICommand but a _LAIResponse
    resp = _LAIResponse.from_str("[S01V14Huber ControlC1")
    print(resp)  # expect '[S01V14Huber ControlC1'
    print(repr(resp))  # expect _LAIResponse(<_LAICommandType.VERIFY: 'V'>, 'Huber Control')
    print("---")

    chiller = HuberChiller()
    chiller.open()
    print("operating_range_limit", chiller.operating_range_limit)
    print("set_point_range_limit", chiller.set_point_range_limit)
    print("set_point_value", chiller.set_point_value)
    print("internal_actual_value", chiller.internal_actual_value)
    print("external_actual_value", chiller.external_actual_value)
    print("temperature_control_mode", chiller.temperature_control_mode)
    print("alarm_status", chiller.alarm_status)
    chiller.set_point_value = 22.0
    chiller.temperature_control_mode = HuberTemperatureControlMode.INTERNAL
    chiller.clear_alarm_status()
    input("Press enter to turn temperature control off")
    chiller.temperature_control_mode = HuberTemperatureControlMode.OFF
    chiller.close()
