"""
Utility classes and functions for the device drivers
"""
from enum import Enum


class StrEnum(str, Enum):
    """Enum where members are also (and must be) strs"""


def int_to_hex(value: int, num_bytes: int = 2) -> str:
    """
    Converts the given integer `value` to a hexadecimal string

    If `value` would result in a hexadecimal number with more than `num_bytes` bytes the higher digits simply get cut
    off.
    """
    return f"{value:0>{num_bytes}X}"[-num_bytes:]


def hex_to_int(value: str, num_bytes: int = 2, is_signed: bool = True) -> int:
    """
    Converts the given hexadecimal `value` to an int
    """
    val = int(value, base=16) if value.startswith(("0x", "0X")) else int("0x" + value, base=16)
    if is_signed and val & (1 << (num_bytes * 8 - 1)) != 0:
        val = val - (1 << num_bytes * 8)
    return val


def hex_to_temperature(value: str, num_bytes: int = 2) -> float:
    """
    Converts the given hexadecimal `value` to a temperature value
    """
    # temperatures are encoded as values from -327.66 °C to 327.67 °C with 0.01 K increments
    return hex_to_int(value, num_bytes) * 0.01


def temperature_to_hex(value: float, num_bytes: int = 4):
    """
    Converts the given temperature `value` to a hexadecimal value
    """
    # temperatures are encoded as values from -327.66 °C to 327.67 °C with 0.01 K increments
    return int_to_hex(int(value * 100), num_bytes)
