"""
An interface for implementing a purification device driver for the CETONI SiLA SDK
"""

from __future__ import annotations

import datetime
import logging
from abc import abstractmethod
from typing import List, NamedTuple

from sila_cetoni.device_driver_abc import DeviceDriverABC

logger = logging.getLogger(__name__)


class Reminder(NamedTuple):
    """
    A simple reminder class
    """

    name: str
    date: datetime.date


class Alert(NamedTuple):
    """
    A simple alert class
    """

    level: str
    category: str
    issue: str
    timestamp: datetime.datetime
    acknowledged: bool
    muted: bool


class PurificationDeviceNotFoundException(Exception):
    """
    An exception that indicates that a purification device is not available (any more)
    """

    def __init__(self, msg: str = "No purification device detected"):
        super().__init__(msg)


class PurificationDeviceInterface(DeviceDriverABC):
    """
    Interface for a purification device driver
    """

    _name: str = ""
    _intake_conductivity: float = 0.0
    _intake_temperature: float = 0.0
    _conductivity: float = 0.0
    _temperature: float = 0.0
    _is_dispensing: bool = False
    _is_dispensing_possible: bool = False
    _bag_tank_capacity: bool = False
    _bag_tank_available_volume: bool = False
    _reminders: List[Reminder] = []
    _alerts: List[Alert] = []
    _input_pressure: float = 0.0
    _rejection_rate: float = 0.0

    def __init__(self):
        super().__init__()

    @abstractmethod
    def start_dispensing(self) -> None:
        """
        Start dispensing of purified water. Depending on the device or the configured dispensing mode dispensing might
        be stopped automatically after a configured period of time (time-controlled mode), after a certain volume has
        been dosed (volume-controlled mode), or dispensing might not be stopped until StopDispensing is called ("manual"
        mode).
        """
        raise NotImplementedError()

    @abstractmethod
    def stop_dispensing(self) -> None:
        """
        Stop dispensing of purified water. Depending on the device or the configured dispensing mode it might be
        possible to not only stop dispensing in "manual" mode but also in volume-controlled or time-controlled mode.
        """
        raise NotImplementedError()

    @property
    def name(self) -> str:
        return self._name

    @property
    def intake_conductivity(self) -> float:
        """
        Conductivity of the water coming from the intake
        """
        return self._intake_conductivity

    @property
    def intake_temperature(self) -> float:
        """
        Temperature of the water coming from the intake
        """
        return self._intake_temperature

    @property
    def conductivity(self) -> float:
        """
        Conductivity of the purified water
        """
        return self._conductivity

    @property
    def temperature(self) -> float:
        """
        Temperature of the purified water
        """
        return self._temperature

    @property
    def is_dispensing(self) -> bool:
        """
        Whether or not the device is currently dispensing purified water
        """
        return self._is_dispensing

    @property
    def is_dispensing_possible(self) -> bool:
        """
        Whether or not it is currently possible to dispense purified water
        """
        return self._is_dispensing_possible

    @property
    def bag_tank_capacity(self) -> bool:
        """
        Total capacity of the bag tank holding the purified water
        """
        return self._bag_tank_capacity

    @property
    def bag_tank_available_volume(self) -> bool:
        """
        Available volume of purified water in the bag tank
        """
        return self._bag_tank_available_volume

    @property
    def reminders(self) -> List[Reminder]:
        """
        A list of upcoming reminders
        """
        return self._reminders

    @property
    def alerts(self) -> List[Alert]:
        """
        Alerts
        """
        return self._alerts

    @property
    def input_pressure(self) -> float:
        """
        `/pre/input_pressure`
        """
        return self._input_pressure

    @property
    def rejection_rate(self) -> float:
        """
        `/pre/ro/rejection_rate`
        """
        return self._rejection_rate
