"""
A device driver implementation for Sartorius Arium I/II water purification devices
"""

from __future__ import annotations

import datetime
import logging
import threading
import time
from enum import Enum
from threading import Event, Thread
from typing import TYPE_CHECKING, Any, List, Optional

import requests
from requests.exceptions import ConnectionError, ConnectTimeout

if TYPE_CHECKING:
    import revpimodio2 as rp2

try:
    import coloredlogs
except ModuleNotFoundError:
    print("Cannot find coloredlogs! Please install coloredlogs, if you'd like to have nicer logging output:")
    print("`pip install coloredlogs`")

from .abc import Alert, PurificationDeviceInterface, PurificationDeviceNotFoundException, Reminder

logger = logging.getLogger(__name__)


class SartoriusAriumNotFoundException(PurificationDeviceNotFoundException):
    def __init__(self, server_url: Optional[str] = None):
        super().__init__(
            f"No Sartorius Arium device available at {server_url if server_url is not None else 'all'}. Check the LAN "
            "connection to the device and ensure that the device running this driver has the correct IP address "
            "(169.254.0.1)"
        )


class StrEnum(str, Enum):
    pass


class _Endpoints(StrEnum):
    SYSTEM_MAINTENANCE_REMINDER = "/system/maintenance/reminder"
    SYSTEM_QUALITY_MODULE_ALERT = "/system/quality_module/alert"
    SYSTEM_WANT_RESET = "/system/want_reset"
    PRE_RO_QUALITY_IN_CONDUCTIVITY = "/pre/ro/quality_in/conductivity"
    PRE_RO_QUALITY_IN_TEMPERATURE = "/pre/ro/quality_in/temperature"
    PRE_RO_QUALITY_OUT_CONDUCTIVITY = "/pre/ro/quality_out/conductivity"
    PRE_RO_QUALITY_OUT_TEMPERATURE = "/pre/ro/quality_out/temperature"
    PRE_RO_REJECTION_RATE = "/pre/ro/rejection_rate"
    PRE_INPUT_PRESSURE = "/pre/input_pressure"
    PRE_RO_MODULE_REMINDER = "/pre/ro_module/reminder"
    PRE_PREFILTER_REMINDER = "/pre/prefilter/reminder"
    PRE_CLEANING_REMINDER = "/pre/cleaning/reminder"
    TANK_BAG_CONTENT = "/tank/bag/content"
    TANK_BAG_REMINDER = "/tank/bag/reminder"
    TANK_PROCESS = "/tank/process"
    TANK_PROCESS_IS_OUTPUT_POSSIBLE = "/tank/process/is_output_possible"
    TANK_PROCESS_VOLUME_AVAILABLE = "/tank/process/volume_available"
    FINAL_QUALITY_IN_CONDUCTIVITY = "/final/quality_in/conductivity"
    FINAL_QUALITY_IN_TEMPERATURE = "/final/quality_in/temperature"
    FINAL_QUALITY_OUT_CONDUCTIVITY = "/final/quality_out/conductivity"
    FINAL_QUALITY_OUT_TEMPERATURE = "/final/quality_out/temperature"
    FINAL_CARTRIDGE_REMINDER = "/final/cartridge/reminder"
    FINAL_UV_LAMP_CURRENT = "/final/uv_lamp/current"
    FINAL_UV_LAMP_REMINDER = "/final/uv_lamp/reminder"
    FINAL_ULTRA_FILTER_REMINDER = "/final/ultra_filter/reminder"
    FINAL_TOC_VALUE = "/final/toc/value"
    FINAL_TOC_ALERT = "/final/toc/alert"
    FINAL_TOC_REMINDER = "/final/toc/reminder"
    FINAL_CLEANING_REMINDER = "/final/cleaning/reminder"
    FINAL_FLOW_SENSOR_ALERT = "/final/flow_sensor/alert"
    FINAL_PROCESS = "/final/process"
    FINAL_PROCESS_IS_OUTPUT_POSSIBLE = "/final/process/is_output_possible"
    FINAL_PROCESS_VOLUME_AVAILABLE = "/final/process/volume_available"
    DISPENSE_FINAL_FILTER_REMINDER = "/dispense/final_filter/reminder"

    def to_name(self) -> str:
        return self[1:].replace("/", "_").upper()

    def friendly_reminder_name(self) -> str:
        """
        In case this endpoint is associated to return a known reminder this function returns the friendly human readable
        name for this reminder, otherwise this function returns the same as `to_name`
        """
        if self == _Endpoints.PRE_CLEANING_REMINDER:
            return "System Cleaning Reminder"
        elif self == _Endpoints.SYSTEM_MAINTENANCE_REMINDER:
            return "System Maintenance Reminder"
        elif self == _Endpoints.TANK_BAG_REMINDER:
            return "Bag Exchange Reminder"
        elif self == _Endpoints.PRE_PREFILTER_REMINDER:
            return "Prefilter Reminder"
        elif self == _Endpoints.PRE_RO_MODULE_REMINDER:
            return "RO Module Reminder"
        elif self == _Endpoints.FINAL_CARTRIDGE_REMINDER:
            return "Cartridge Reminder"
        elif self == _Endpoints.FINAL_UV_LAMP_REMINDER:
            return "UV-Lamp Reminder"
        elif self == _Endpoints.DISPENSE_FINAL_FILTER_REMINDER:
            return "Endfilter Reminder"  # not confirmed
        elif self == _Endpoints.FINAL_TOC_REMINDER:
            return "TOC Reminder"  # not confirmed
        elif self == _Endpoints.PRE_INPUT_PRESSURE:
            return "Input Pressure"
        else:
            return self.to_name()


_LOCAL_TIMEZONE = datetime.datetime.now().astimezone().tzinfo
_DEFAULT_DATE = {"year": 1, "month": 1, "day": 1}
_DEFAULT_TIME = {"hour": 0, "minute": 0}


class SartoriusAriumBase(PurificationDeviceInterface):
    """
    The Sartorius Arium water purification device driver implementation
    """

    _server_url: str
    _trigger_output: str

    __requested_endpoints: List[_Endpoints]
    __alerts: List[Alert]
    __reminders: List[Reminder]

    _stop_event: Event

    def __init__(self, server_url: str, trigger_output: str):
        """
        Parameters
        ----------
        server_url: str
            The URL of the REST server to use for querying device parameters and status info (e.g.
            "http://169.254.0.100/rest/v1_0")
        trigger_output: str
            The output pin to which the trigger port of the device is connected to trigger dosing of purified water
        """
        super().__init__()

        self._server_url = server_url
        self._trigger_output = trigger_output

        self.__requested_endpoints = []
        self.__reminders = []
        self.__alerts = []

        self._stop_event = Event()

        try:
            resp = self.__get(_Endpoints.FINAL_PROCESS)
            self._name = resp.get("product_type", None)
            logger.info(f"product type {self._name!r}")
            if self._name is None or self._name != "Arium Comfort":
                raise SartoriusAriumNotFoundException(self._server_url)
        except (ConnectionError, ConnectTimeout) as err:
            logger.warning(err, exc_info=err)
            raise SartoriusAriumNotFoundException(self._server_url)

        def poll():
            while not self._stop_event.wait(1):
                self.__requested_endpoints = []
                self.__reminders = []
                self.__alerts = []

                try:
                    intake_conductivity = self.__get(_Endpoints.PRE_RO_QUALITY_OUT_CONDUCTIVITY)
                    self._intake_conductivity = intake_conductivity.get("value", None) or 0.0
                    intake_temperature = self.__get(_Endpoints.PRE_RO_QUALITY_OUT_TEMPERATURE)
                    self._intake_temperature = intake_temperature.get("value", None) or 0.0

                    conductivity = self.__get(_Endpoints.FINAL_QUALITY_OUT_CONDUCTIVITY)
                    self._conductivity = conductivity.get("value", None) or 0.0
                    temperature = self.__get(_Endpoints.FINAL_QUALITY_OUT_TEMPERATURE)
                    self._temperature = temperature.get("value", None) or 0.0

                    self._get_is_dispensing()

                    tank_output_possible = self.__get(_Endpoints.TANK_PROCESS_IS_OUTPUT_POSSIBLE)
                    final_output_possible = self.__get(_Endpoints.FINAL_PROCESS_IS_OUTPUT_POSSIBLE)
                    logger.debug(
                        f"tank_output_possible {tank_output_possible!r}, "
                        f"final_output_possible {final_output_possible!r}"
                    )
                    self._is_dispensing_possible = final_output_possible

                    tank_bag_content = self.__get(_Endpoints.TANK_BAG_CONTENT)
                    self._bag_tank_capacity = tank_bag_content.get("volume", {}).get("value", None) or 0.0
                    tank_volume_avail = self.__get(_Endpoints.TANK_PROCESS_VOLUME_AVAILABLE)
                    final_volume_avail = self.__get(_Endpoints.FINAL_PROCESS_VOLUME_AVAILABLE)
                    logger.debug(
                        f"tank_volume_avail {tank_volume_avail.get('value', None)!r}, "
                        f"final_volume_avail {final_volume_avail.get('value', None)!r}"
                    )
                    self._bag_tank_available_volume = final_volume_avail.get("value", None) or 0.0

                    input_pressure = self.__get(_Endpoints.PRE_INPUT_PRESSURE)
                    self._input_pressure = input_pressure.get("value", None) or 0.0
                    rejection_rate = self.__get(_Endpoints.PRE_RO_REJECTION_RATE)
                    self._rejection_rate = rejection_rate.get("value", None) or 0.0

                    for endpoint in set(_Endpoints) - set(self.__requested_endpoints):
                        self.__get(endpoint)
                    self._reminders = sorted(self.__reminders, key=lambda r: r.date)
                    self._alerts = sorted(self.__alerts, key=lambda r: r.timestamp)
                except requests.exceptions.ConnectTimeout as err:
                    logger.warning(err, exc_info=err)
                except:
                    if self._stop_event.is_set():
                        break
                    raise

        # start polling for value changes
        Thread(target=poll, name="PollingThread").start()

    def _get_is_dispensing(self) -> bool:
        """
        Requests the `TANK_PROCESS` and `FINAL_PROCESS` endpoints from the REST API and sets the `_is_dispensing` member

        Returns
        -------
        bool:
            The value that `_is_dispensing` was set to
        """
        tank_process = self.__get(_Endpoints.TANK_PROCESS)
        final_process = self.__get(_Endpoints.FINAL_PROCESS)
        logger.debug(
            f"tank_process {tank_process.get('process')!r} {tank_process.get('state')!r}, "
            f"final_process {final_process.get('process')!r} {final_process.get('state')!r}"
        )
        self._is_dispensing = (
            final_process.get("process", "") == "DISPENSE" and final_process.get("state", "") == "RUNNING"
        )
        return self._is_dispensing

    def __get(self, endpoint: _Endpoints) -> Any:
        """
        Performs a GET request to the REST API of the Arium and appends potential a alert/reminder to the current list
        of alerts/reminders

            :param endpoint: The endpoint to perform the GET for

            :returns: The received JSON response or `None` if the request was not successful
        """
        if self._stop_event.is_set():
            return
        resp = requests.get(self._server_url + endpoint, timeout=2)
        self.__requested_endpoints.append(endpoint)
        logger.debug(f"GET {endpoint} response {resp} ({resp.json() if resp.ok else None})")
        # return resp.json() if resp.ok else None
        if resp.ok:
            json = resp.json()
            self.__append_alert_reminder(json, endpoint.friendly_reminder_name())
            return json

    def __append_alert_reminder(self, response: Any, name: Optional[str] = None):
        """
        Takes the alert or reminder (if present) from the given `response` and appends them to the current list of
        alerts/reminders

            :param response: The response from the REST API
            :param name: An optional name for the reminder (if present in `response`)
                         This name will take precedence over the name found in the response body. If no name is given in
                         this parameter an there is no name in the response body the name will be set to "unknown"
        """
        if not isinstance(response, dict):
            return

        alert = response.get("alert", None)
        if isinstance(alert, dict):
            self.__alerts.append(
                Alert(
                    alert.get("level", "unknown"),
                    alert.get("category", "unknown"),
                    (f"{name} " if name else "") + alert.get("issue", "unknown"),
                    datetime.datetime(
                        **(alert.get("date", None) or _DEFAULT_DATE),
                        **(alert.get("time", None) or _DEFAULT_TIME),
                        tzinfo=_LOCAL_TIMEZONE,
                    ),
                    alert.get("acknowledged", False),
                    alert.get("muted", False),
                )
            )

        date = response.get("date", None)
        if date is not None:
            name = name or response.get("consumable_type", None) or "unknown"
            self.__reminders.append(Reminder(name, datetime.date(**date)))

    def start(self):
        # This device does not need to be started - that's handled in `__init__()`
        pass

    def stop(self):
        self._stop_event.set()


class SartoriusArium(SartoriusAriumBase):

    __rev_pi: rp2.RevPiModIO

    def __init__(self, server_url: str, trigger_output: str):
        super().__init__(server_url, trigger_output)

        import revpimodio2 as rp2

        self.__rev_pi = rp2.RevPiModIO(autorefresh=True, shared_procimg=True)
        self.__rev_pi.cycletime = 40  # the default 20 ms cannot be held
        self.__rev_pi.mainloop(blocking=False)

    def __trigger_foot_switch(self) -> None:
        """
        Triggers the foot switch of the Arium device which signals starting/stopping of dispensing to the device
        """
        self.__rev_pi.io[self._trigger_output].value = True
        time.sleep(0.5)
        self.__rev_pi.io[self._trigger_output].value = False

    def start_dispensing(self) -> None:
        major_error_alerts = list(filter(lambda a: a.level == "MAJOR_ERROR", self.alerts))
        if major_error_alerts:
            raise RuntimeError(
                f"Could not start dispensing because the device has MAJOR_ERROR level alerts: "
                f"{[f'{a.issue!r} occurred at {a.timestamp!s}' for a in major_error_alerts]}"
            )

        if not self.is_dispensing_possible:
            raise RuntimeError("Cannot start dispensing because the device is not ready")

        if not self.is_dispensing:
            self.__trigger_foot_switch()
            if not self._get_is_dispensing():
                raise RuntimeError("Device did not start dispensing")

    def stop_dispensing(self) -> None:
        if self.is_dispensing:
            self.__trigger_foot_switch()

    def stop(self):
        super().stop()
        self.__rev_pi.exit()


class SartoriusAriumSim(SartoriusAriumBase):

    __port: int = 5000
    __sim_server_app_thread: threading.Thread

    def __init__(self, server_url: str, trigger_output: str):
        from ..sim.sartorius_arium import app

        logging.getLogger("werkzeug").setLevel(logging.WARNING)

        self.__sim_server_app_thread = threading.Thread(
            target=app.run,
            kwargs={"port": self.__port, "use_reloader": False, "threaded": False, "processes": 0},
            daemon=True,
        )
        self.__sim_server_app_thread.start()
        time.sleep(0.5)

        server_url_parts = server_url.split("/")
        server_url_parts[2] = f"localhost:{self.__port}"
        self.__port += 1

        super().__init__("/".join(server_url_parts), trigger_output)

    def start_dispensing(self) -> None:
        major_error_alerts = list(filter(lambda a: a.level == "MAJOR_ERROR", self.alerts))
        if major_error_alerts:
            raise RuntimeError(
                f"Could not start dispensing because the device has MAJOR_ERROR level alerts: "
                f"{[f'{a.issue!r} occurred at {a.timestamp!s}' for a in major_error_alerts]}"
            )

        if not self.is_dispensing_possible:
            raise RuntimeError("Cannot start dispensing because the device is not ready")

        if not self.is_dispensing:
            self._is_dispensing = True

    def stop_dispensing(self) -> None:
        if self.is_dispensing:
            self._is_dispensing = False


# ----------------------------------------------------------------------------
# test
if __name__ == "__main__":
    LOGGING_FORMAT = "%(asctime)s [%(threadName)-12.12s] %(levelname)-8s| %(name)s %(module)s.%(funcName)s: %(message)s"
    logging_level = logging.INFO
    try:
        coloredlogs.install(fmt=LOGGING_FORMAT, datefmt="%Y-%m-%d %H:%M:%S,%f", level=logging_level)
    except NameError:
        logging.basicConfig(format=LOGGING_FORMAT, level=logging_level)

    arium = SartoriusArium()
    time.sleep(1)  # wait for the actual values
    print("intake_conductivity", arium.intake_conductivity)
    print("intake_temperature", arium.intake_temperature)
    print("conductivity", arium.conductivity)
    print("temperature", arium.temperature)
    print("is_dispensing", arium.is_dispensing)
    print("is_dispensing_possible", arium.is_dispensing_possible)
    print("bag_tank_capacity", arium.bag_tank_capacity)
    print("bag_tank_available_volume", arium.bag_tank_available_volume)
    print("reminders", arium.reminders)
    print("alerts", arium.alerts)
    print("input_pressure", arium.input_pressure)
    print("rejection_rate", arium.rejection_rate)
    print("start_dispensing")
    arium.start_dispensing()
    print("waiting 2 seconds...")
    time.sleep(2)
    print("stop_dispensing")
    arium.stop_dispensing()
    print("intake_conductivity", arium.intake_conductivity)
    print("intake_temperature", arium.intake_temperature)
    print("conductivity", arium.conductivity)
    print("temperature", arium.temperature)
    print("is_dispensing", arium.is_dispensing)
    print("is_dispensing_possible", arium.is_dispensing_possible)
    print("bag_tank_capacity", arium.bag_tank_capacity)
    print("bag_tank_available_volume", arium.bag_tank_available_volume)
    print("reminders", arium.reminders)
    print("alerts", arium.alerts)
    print("input_pressure", arium.input_pressure)
    print("rejection_rate", arium.rejection_rate)
    arium.stop()
