"""
An interface for implementing a stirring device driver for the CETONI SiLA SDK

Author
------
Florian Meinicke (florian.meinicke@cetoni.de)

Date
----
21.07.2022
"""

from __future__ import annotations

import logging
import time
import traceback
from abc import abstractmethod, abstractstaticmethod
from typing import Optional

import serial
import serial.threaded
import serial.tools.list_ports
from sila_cetoni.device_driver_abc import DeviceDriverABC

logger = logging.getLogger(__name__)


class _ReaderThread(serial.threaded.ReaderThread):
    """
    Implements a serial port read loop and dispatches to a Protocol instance but do it with threads.

    Provides the connection from the `SerialStirringDeviceInterface` instance (which creates an instance of this
    `_ReaderThread` class) to the protocol instance (`_SerialStirringDeviceReader`) that is created by this
    `_ReaderThread` so that the protocol instance that receives the data from the serial port can set the actual value
    of the stirring device in the `SerialStirringDeviceInterface` instance
    """

    stirring_device: SerialStirringDeviceInterface = None

    def __init__(self, serial_instance, protocol_factory, stirring_device):
        super().__init__(serial_instance, protocol_factory)
        self.stirring_device: SerialStirringDeviceInterface = stirring_device


class _SerialStirringDeviceReader(serial.threaded.LineReader):
    """
     Reads/writes to/from a stirring device via a serial connection

     The `SerialStirringDeviceInterface` instance that is represented is obtained from the transport object (a
    `_ReaderThread`) that creates an instance of this reader class
    """

    __stirring_device: SerialStirringDeviceInterface = None

    # implements serial.threaded.LineReader ----------------------------------
    def connection_made(self, transport: _ReaderThread):
        super().connection_made(transport)

        self.__stirring_device: SerialStirringDeviceInterface = transport.stirring_device
        logger.debug("port opened")

        # validate that we have the correct stirring device
        self.write_line(self.__stirring_device.unique_stirring_device_identifier_request)
        time.sleep(1.5)
        buffer: bytes = transport.serial.read_all()
        logger.debug(f"buffer {buffer}")
        if not buffer or not self.__stirring_device.is_valid_stirring_device(buffer):
            raise self.__stirring_device.not_found_exception()

    def handle_line(self, data: str):
        logger.debug("line received: {}".format(repr(data)))

        self.__stirring_device.on_serial_data_received(data)

    def connection_lost(self, exc):
        if exc:
            traceback.print_exception(exc, exc, None)
        logger.debug("port closed")


class StirringDeviceInterface(DeviceDriverABC):
    """
    Interface for a stirring device driver
    """

    _is_stirring: bool
    _rpm: float
    _power: float

    def __init__(self):
        super().__init__()
        self._is_stirring = False
        self._rpm = 0
        self._power = 0

    @property
    def is_stirring(self) -> float:
        """
        Returns whether the device is currently stirring or not
        """
        return self._is_stirring

    @property
    def rpm(self) -> float:
        """
        Returns the stirring speed in RPM
        """
        return self._rpm

    @rpm.setter
    @abstractmethod
    def rpm(self, rpm: float) -> None:
        """
        Set the stirring power
        """
        raise NotImplementedError()

    @property
    def power(self) -> float:
        """
        Returns the stirring power in percent
        """
        return self._power

    @power.setter
    @abstractmethod
    def power(self, power: float) -> None:
        """
        Set the stirring power
        """
        raise NotImplementedError()

    @abstractmethod
    def start_stirring(self) -> None:
        """
        Starts stirring
        """
        raise NotImplementedError()

    @abstractmethod
    def stop_stirring(self) -> None:
        """
        Stops stirring
        """
        raise NotImplementedError()


class StirringDeviceNotFoundException(Exception):
    """
    An exception that indicates that a stirring device is not available (any more) over a serial connection
    """

    def __init__(self, port: Optional[str] = None, msg: str = "No stirring device detected on {port}"):
        super().__init__(msg.format(port=port if port is not None else "any serial port"))


class SerialStirringDeviceInterface(StirringDeviceInterface):
    """
    Interface for a stirring device device driver that uses a serial communication protocol
    """

    # You can derive from `StirringDeviceNotFoundException` to give a more detailed error
    # message. Set `not_found_exception = YourStirringDeviceNotFoundException` in your
    # custom StirringDevice class
    not_found_exception = StirringDeviceNotFoundException
    # The delimiter that is sent at the end of every message
    delimiter: str = "\r"

    __serial: serial.Serial
    __reader_thread: serial.threaded.ReaderThread
    _protocol: _SerialStirringDeviceReader

    __COMMON_SERIAL_ARGS = {
        "baudrate": 9600,
        "bytesize": serial.EIGHTBITS,
        "parity": serial.PARITY_NONE,
        "stopbits": serial.STOPBITS_ONE,
    }

    def __init__(self, port: Optional[str] = None):
        """
        :param port: (optional) The serial port of the stirring device
        """
        super().__init__()

        try:
            self.__serial = serial.Serial(port, **self.__COMMON_SERIAL_ARGS)
        except serial.SerialException:
            raise self.not_found_exception(port)
        self.__reader_thread = _ReaderThread(self.__serial, _SerialStirringDeviceReader, self)
        self.__reader_thread.name = "SerialStirringDeviceReaderThread"

    def _autodetect_serial_port(self):
        """
        Goes through all available serial ports of the system and tries to find the first port connected to a stirring
        device
        """

        for info in serial.tools.list_ports.comports():
            logger.debug(f"autodetection trying port {info.device}")
            try:
                ser = serial.Serial(info.device, **self.__COMMON_SERIAL_ARGS, timeout=2, write_timeout=2)
                # read something from the stirring device that uniquely identifies it
                TERMINATOR = _SerialStirringDeviceReader.TERMINATOR
                s = f"{self.unique_stirring_device_identifier_request}".encode("utf-8") + TERMINATOR
                logger.debug(f"send {s}")
                ser.write(f"{self.unique_stirring_device_identifier_request}".encode("utf-8") + TERMINATOR)
                time.sleep(0.1)
                if self.is_valid_stirring_device(ser.read_all()):
                    logger.info(f"stirring device detected on serial port {ser.port}")
                    self.__serial.port = ser.port
                    break
            except (TimeoutError, serial.SerialTimeoutException, serial.SerialException):
                continue
        else:
            raise self.not_found_exception()

    def open(self, port: str = ""):
        """
        Connects to the stirring device via the serial port with the given `port`

        :param port: (optional) The serial port of the stirring device. If not given, tries to autodetect a stirring
                     device by trying all available serial ports until a valid port is found
        """
        if self.__serial.isOpen():
            return

        if not port:
            self._autodetect_serial_port()
        else:
            self.__serial.port = port

        self.__serial.open()

    def start(self):
        if not self.__serial.isOpen():
            raise self.not_found_exception(self.port)
        try:
            self.__reader_thread.start()
            self._protocol: _SerialStirringDeviceReader = self.__reader_thread.connect()[1]
        except RuntimeError:
            raise self.not_found_exception(self.port)

    def stop(self):
        self.close()

    def close(self):
        """
        Closes the communication to the stirring device
        """
        if self.__reader_thread.is_alive():
            self.__reader_thread.close()
        self.__serial.close()
        logger.debug("stop done")

    @property
    def port(self) -> str:
        """
        Returns the serial port through which this device driver communicates with the stirring device
        """
        return self.__serial.port

    @abstractmethod
    def on_serial_data_received(self, data: str) -> None:
        """
        Function to handle the `data` read from the stirring device
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def unique_stirring_device_identifier_request(self) -> str:
        """
        This function should return a string containing a request for the stirring device to return some kind of
        unique identifier or something that can be used to identify if the current serial connection uses the correct
        stirring device (e.g. this could be a request to return the serial number of the stirring device).
        The response of the device will be passed to the `is_valid_stirring_device()` function to validate it.
        """
        raise NotImplementedError()

    @abstractstaticmethod
    def is_valid_stirring_device(data: bytes) -> bool:
        """
        This function should return whether `data` indicates that the current serial connection is using the correct
        stirring device.

        :param data: The serial data which was read as a response to writing the
                     `unique_stirring_device_identifier_request`
        """
        raise NotImplementedError()
