# Generated by sila2.code_generator; sila2.__version__: 0.8.3
from __future__ import annotations
import logging

import math
import time
from concurrent.futures import Executor
from queue import Queue
from threading import Event
from typing import Optional

from sila2.server import MetadataDict, SilaServer
from sila_cetoni.application.server_configuration import ServerConfiguration
from sila_cetoni.application.system import ApplicationSystem

from ....device_drivers.abc import StirringDeviceInterface
from ....device_drivers.twomag_mixdrive import MIXdriveError
from ..generated.stirringservice import (
    RPM,
    DeviceInManualMode,
    ParameterOutOfRange,
    Percent,
    SetPower_Responses,
    SetRPM_Responses,
    StartStirring_Responses,
    StirringServiceBase,
    StopStirring_Responses,
)

logger = logging.getLogger(__name__)

class StirringServiceImpl(StirringServiceBase):
    __device: StirringDeviceInterface
    __config: ServerConfiguration
    __stop_event: Event

    def __init__(self, parent_server: SilaServer, device: StirringDeviceInterface, executor: Executor):
        super().__init__(parent_server)
        self.__device = device
        self.__stop_event = Event()

        def update_is_stirring(stop_event: Event):
            new_is_stirring = is_stirring = self.__device.is_stirring
            while not stop_event.is_set():
                new_is_stirring = self.__device.is_stirring
                if new_is_stirring != is_stirring:
                    is_stirring = new_is_stirring
                    self.update_IsStirring(is_stirring)
                time.sleep(1)

        def update_rpm(stop_event: Event):
            new_rpm = rpm = self.__device.rpm
            while not stop_event.is_set():
                new_rpm = self.__device.rpm
                if not math.isclose(new_rpm, rpm):
                    rpm = new_rpm
                    self.update_RPM(rpm)
                time.sleep(1)

        def update_power(stop_event: Event):
            new_power = power = self.__device.power
            while not stop_event.is_set():
                new_power = self.__device.power
                if not math.isclose(new_power, power):
                    power = new_power
                    self.update_Power(power)
                time.sleep(1)

        executor.submit(update_is_stirring, self.__stop_event)
        executor.submit(update_rpm, self.__stop_event)
        executor.submit(update_power, self.__stop_event)

    def start(self) -> None:
        self.__config = ServerConfiguration(self.parent_server.server_name, ApplicationSystem().device_config.name)
        # restore last values
        if self.__config.stirring_rpm is not None:
            try:
                self.__device.rpm = self.__config.stirring_rpm
            except MIXdriveError as err:
                logger.warning(f"Failed to set last RPM to {self.__config.stirring_rpm}: {err}", exc_info=err)
        if self.__config.stirring_power is not None:
            try:
                self.__device.power = self.__config.stirring_power
            except MIXdriveError as err:
                logger.warning(f"Failed to set last power to {self.__config.stirring_power}: {err}", exc_info=err)

        # initial values
        self.update_IsStirring(self.__device.is_stirring)
        self.update_RPM(self.__device.rpm)
        self.update_Power(self.__device.power)
        super().start()

    def stop(self) -> None:
        super().stop()
        self.__stop_event.set()
        self.__config.write()

    def update_RPM(self, rpm: RPM, queue: Optional[Queue[RPM]] = None):
        self.__config.stirring_rpm = rpm
        super().update_RPM(rpm, queue)

    def update_Power(self, power: Percent, queue: Optional[Queue[Percent]] = None):
        self.__config.stirring_power = power
        super().update_Power(power, queue)

    def StartStirring(self, *, metadata: MetadataDict) -> StartStirring_Responses:
        try:
            self.__device.start_stirring()
        except MIXdriveError as err:
            if err.is_manual_mode_error:
                raise DeviceInManualMode(err.msg + err.extra_msg)
            raise

    def StopStirring(self, *, metadata: MetadataDict) -> StopStirring_Responses:
        try:
            self.__device.stop_stirring()
        except MIXdriveError as err:
            if err.is_manual_mode_error:
                raise DeviceInManualMode(err.msg + err.extra_msg)
            raise

    def SetRPM(self, value: RPM, *, metadata: MetadataDict) -> SetRPM_Responses:
        try:
            self.__device.rpm = value
        except MIXdriveError as err:
            if err.is_out_of_range_error:
                raise ParameterOutOfRange(err.msg + f", invalid value {value}" + err.extra_msg)
            raise

    def SetPower(self, power: Percent, *, metadata: MetadataDict) -> SetPower_Responses:
        try:
            self.__device.power = power
        except MIXdriveError as err:
            if err.is_out_of_range_error:
                raise ParameterOutOfRange(err.msg + f", invalid value {power}" + err.extra_msg)
            raise
